"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Request = void 0;
const https_1 = require("https");
const http_1 = require("http");
const querystring_1 = require("querystring");
const url_1 = require("url");
const stream_1 = require("stream");
const zlib_1 = require("zlib");
const base_client_1 = require("./base-client");
const request_error_1 = require("./request-error");
const protos = { http: http_1.request, https: https_1.request }, ports = { http: 80, https: 443 };
class Request extends base_client_1.BaseClient {
    constructor(method, url) {
        super();
        this._body = '';
        const parsed = (0, url_1.parse)(url);
        this._params = { method };
        this.location(url);
        this._headers['Accept'] = 'application/json';
        this._headers['Accept-Encoding'] = 'gzip, deflate, br';
    }
    get method() {
        return this._params.method;
    }
    get path() {
        return this._params.path;
    }
    get host() {
        return this._params.host;
    }
    get querystring() {
        return (0, querystring_1.stringify)(Object.assign({}, this._query));
    }
    location(url) {
        const parsed = (0, url_1.parse)(url);
        this._protocol = (parsed.protocol || 'http:').slice(0, -1);
        this._query = (0, querystring_1.parse)(parsed.query);
        this._params.host = parsed.hostname;
        this._params.port = parsed.port || ports[this._protocol];
        this._params.path = parsed.pathname;
        return this;
    }
    query(arg) {
        let inner;
        if (typeof arg === 'string') {
            inner = (0, querystring_1.parse)(arg);
        }
        else {
            inner = arg;
        }
        Object.keys(inner).forEach((k) => this._query[k] = inner[k]);
        return this;
    }
    cert(certstr) {
        this._params.cert = certstr;
        return this;
    }
    key(keystr) {
        this._params.key = keystr;
        return this;
    }
    ca(castrs) {
        this._params.ca = castrs;
        return this;
    }
    send(body) {
        if (typeof body === 'string') {
            this._body = body;
        }
        else {
            this._body = JSON.stringify(body);
            this._headers['Content-Type'] = 'application/json';
        }
        this._headers['Content-Length'] = Buffer.byteLength(this._body);
        return this;
    }
    attempt() {
        let req;
        const params = Object.assign({}, this._params);
        if (Object.keys(this._query).length) {
            params.path += '?' + (0, querystring_1.stringify)(Object.assign({}, this._query));
        }
        params.headers = Object.assign({}, this._headers);
        return new Promise((resolve, reject) => {
            let res;
            req = protos[this._protocol](params, (response) => {
                res = response;
                res.text = '';
                const output = new stream_1.Writable();
                const onError = (err) => {
                    if (err)
                        console.log(err);
                    output.end();
                };
                output._write = (chunk, encoding, next) => {
                    res.text += chunk.toString();
                    next();
                };
                output.on('close', () => {
                    try {
                        res.body = JSON.parse(res.text);
                    }
                    catch (e) {
                        res.body = null;
                    }
                    if (res.statusCode >= 300) {
                        const rErr = new request_error_1.RequestError('Bad response from server.', this, params, res);
                        return reject(rErr);
                    }
                    resolve(res);
                });
                switch (res.headers['content-encoding']) {
                    case 'br':
                        (0, stream_1.pipeline)(res, (0, zlib_1.createBrotliDecompress)(), output, onError);
                        break;
                    case 'gzip':
                    case 'deflate':
                        (0, stream_1.pipeline)(res, (0, zlib_1.createUnzip)(), output, onError);
                        break;
                    default:
                        (0, stream_1.pipeline)(res, output, onError);
                        break;
                }
            });
            if (this._body.length)
                req.write(this._body);
            req.on('error', (err) => {
                const rErr = new request_error_1.ConnectionError('Connection Error: ' + err.message, this, params);
                reject(rErr);
            });
            this.abort = () => req.abort();
            req.on('abort', () => {
                const rErr = new request_error_1.AbortError(`Request was aborted`, this, params);
                reject(rErr);
            });
            req.end();
        });
    }
    async promise() {
        while (true) {
            await this.runPlugins('onRequest');
            try {
                const res = await this.attempt();
                const pres = await this.runPlugins('onResponse', res);
                if (pres === null || pres === void 0 ? void 0 : pres.retry) {
                    continue;
                }
                return res;
            }
            catch (err) {
                const pres = await this.runPlugins('onError', err);
                if (pres === null || pres === void 0 ? void 0 : pres.retry) {
                    continue;
                }
                throw err;
            }
        }
    }
    async runPlugins(name, ...args) {
        if (!this._pluginInstances)
            this._pluginInstances = this._plugins.map(pc => new pc());
        const plugins = this._pluginInstances.filter(p => Boolean(p[name]));
        let result;
        for (const plugin of plugins) {
            const pres = await plugin[name](this, ...args);
            if (pres === null || pres === void 0 ? void 0 : pres.retry) {
                result = pres;
            }
        }
        return result;
    }
    /*
     * Promise-like functions
     */
    then(resolve, reject) {
        return this.promise().then(resolve, reject);
    }
    catch(cb) {
        return this.promise().catch(cb);
    }
    finally(cb) {
        return this.promise().finally(cb);
    }
    /*
     * Callback support
     */
    end(done) {
        this.then((res) => done(null, res), (err) => done(err));
    }
}
exports.Request = Request;
